// Copyright (c) 2024 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#pragma once
#include "dawn.hpp"
#include "event/Event.hpp"

namespace Dawn {
  class Scene;

  class SceneItem : public std::enable_shared_from_this<SceneItem> {
    private:
      std::weak_ptr<SceneItem> parent;
      std::weak_ptr<Scene> scene;
      bool_t isDisposing = false;

    protected:
      std::vector<std::shared_ptr<SceneItem>> items;

    public:
      std::string name;
      Event<> onEnter;
      Event<> onLeave;

      /**
       * Creates a new SceneItem.
       * 
       * @param parent The parent SceneItem, or nullptr if root (unlikely).
       * @param scene The Scene this SceneItem belongs to.
       */
      SceneItem(
        std::weak_ptr<SceneItem> parent,
        std::weak_ptr<Scene> scene
      );

      /**
       * Adds a SceneItem as a child of this SceneItem.
       * 
       * @tparam SceneItem type to add.
       * @return Pointer to the newly created SceneItem.
       */
      template<typename C>
      std::shared_ptr<C> addSceneItem(const std::string &name = "") {
        auto item = std::make_shared<C>(weak_from_this(), scene);
        auto asItem = std::dynamic_pointer_cast<SceneItem>(item);
        if(asItem == nullptr) {
          throw std::runtime_error("SceneItem must be a SceneItem.");
        }
        asItem->name = name;
        items.push_back(asItem);

        return item;
      }

      /**
       * Finds a SceneItem by name.
       * 
       * @tparam The type of SceneItem to find.
       * @param name The name of the SceneItem to find.
       * @return The SceneItem, or nullptr if not found.
       */
      template<typename C>
      std::shared_ptr<C> find(const std::string &name, const bool_t &recursive = false){
        for(auto &item : items) {
          if(item->name == name) {
            auto casted = std::dynamic_pointer_cast<C>(item);
            if(!casted) throw std::runtime_error("Invalid SceneItem type.");
            return casted;
          }

          if(recursive) {
            auto childFound = item->find<C>(name, recursive);
            if(childFound) return childFound;
          }
        }
        
        return nullptr;
      }

      /**
       * Removes a SceneItem from this SceneItem.
       * 
       * @param item The SceneItem to remove.
       */
      void removeSceneItem(std::shared_ptr<SceneItem> item);

      /**
       * Gets the parent SceneItem.
       * 
       * @return The parent SceneItem.
       */
      std::shared_ptr<SceneItem> getParent();

      /**
       * Gets the Scene this SceneItem belongs to.
       * 
       * @return The Scene this SceneItem belongs to.
       */
      std::shared_ptr<Scene> getScene();

      /**
       * Destroys this SceneItem.
       */
      virtual ~SceneItem();

      friend class Scene;
  };
}