// Copyright (c) 2024 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#include "DawnGLFWHost.hpp"
#include "error/error.hpp"
#include "locale/Language.hpp"
#include "error/erroropengl.hpp"

using namespace Dawn;

DawnGLFWHost::DawnGLFWHost() {

}

DawnGLFWInitResult DawnGLFWHost::init(
  const int32_t argc, const char_t **argv
) {
  if(!glfwInit()) {
    errorLog("GLFW failed to initialize.");
    errorMessageBox(Language::get("glfw.error.glfw_init"));
    return DawnGLFWInitResult::GLFW_INIT_FAILED;
  }

  // Setup window hints
  glfwWindowHint(GLFW_OPENGL_DEBUG_CONTEXT, false);
  glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3);
  glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 3);
  glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);
  glfwWindowHint(GLFW_OPENGL_FORWARD_COMPAT, GL_TRUE);

  // Create window
  this->window = glfwCreateWindow(
    DAWN_GLFW_WIDTH_DEFAULT,
    DAWN_GLFW_HEIGHT_DEFAULT,
    DAWN_GAME_NAME,
    NULL,
    NULL
  );

  if(this->window == nullptr) {
    errorLog("GLFW failed to create window.");
    errorMessageBox(Language::get("glfw.error.window_create"));
    glfwTerminate();
    return DawnGLFWInitResult::WINDOW_CREATE_FAILED;
  }

  // Set back pointer
  glfwSetWindowUserPointer(this->window, this);

  // Load GLAD
  glfwMakeContextCurrent(this->window);
  glfwSwapInterval(1);
  gladLoadGLLoader((GLADloadproc)glfwGetProcAddress);
  try {
    errorCheckOpenGL();
  } catch(const std::runtime_error &e) {
    errorLog(e.what());
    errorMessageBox(Language::get("glfw.error.opengl_error"));
    glfwTerminate();
    return DawnGLFWInitResult::GLAD_GL_CHECK_FAILED;
  }

  // Override the defaults
  int32_t fbWidth, fbHeight;
  glfwGetFramebufferSize(this->window, &fbWidth, &fbHeight);
  glfwGetWindowSize(this->window, &this->windowWidth, &this->windowHeight);
  try {
    errorCheckOpenGL();
  } catch(const std::runtime_error &e) {
    errorLog(e.what());
    errorMessageBox(Language::get("glfw.error.opengl_error"));
    glfwTerminate();
    return DawnGLFWInitResult::GLAD_GL_CHECK_FAILED;
  }
  
  if(
    fbWidth == 0 ||
    fbHeight == 0 ||
    this->windowWidth == 0 ||
    this->windowHeight == 0
  ) {
    errorLog(
      "GLFW Returned incorrect size: " +
      std::to_string(fbWidth) + "x" + std::to_string(fbHeight) + " / " +
      std::to_string(windowWidth) + "x" + std::to_string(windowHeight)
    );
    errorMessageBox(Language::get("glfw.error.viewport_size"));
    glfwTerminate();
    return DawnGLFWInitResult::VIEWPORT_SIZE_FAILED;
  }

  // Init Game
  this->game = std::make_shared<Game>();
  this->game->init();

  // Backbuffer
  backBuffer = std::make_shared<BackBuffer>();
  backBuffer->setSize((float_t)fbWidth, (float_t)fbHeight);

  // Set up callbacks
  glfwSetFramebufferSizeCallback(this->window, [](
    GLFWwindow *window,
    int32_t width,
    int32_t height
  ) {
    auto host = (DawnGLFWHost*)glfwGetWindowUserPointer(window);
    if(host == nullptr) return;
    host->windowWidth = width;
    host->windowHeight = height;
    host->backBuffer->setSize((float_t)width, (float_t)height);
  });

  return DawnGLFWInitResult::SUCCESS;
}

void DawnGLFWHost::start() {
  double_t time, newTime;
  float_t fDelta;
  int32_t updateResult;

  // Main Render Loop 
  time = 0.0f;
  while(!glfwWindowShouldClose(this->window)) {
    // Determine the delta.
    newTime = glfwGetTime();
    fDelta = (float_t)(newTime - time);
    time = newTime;

    // Update the game
    this->game->update();

    // Update GLFW
    glfwSwapBuffers(this->window);
    glfwPollEvents();
  }

  this->game = nullptr;
}

DawnGLFWHost::~DawnGLFWHost() {
  this->backBuffer = nullptr;
  this->game = nullptr;

  // Terminate GLFW
  if(this->window != nullptr) {
    glfwDestroyWindow(this->window);
    glfwTerminate();
  }
}