// Copyright (c) 2023 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#pragma once
#include "dawn.hpp"
#include "dawnopengl.hpp"

namespace Dawn {
  enum MeshDrawMode {
    TRIANGLES = GL_TRIANGLES,
    TRIANGLE_STRIP = GL_TRIANGLE_STRIP,
    TRIANGLE_FAN = GL_TRIANGLE_FAN,
    LINES = GL_LINES,
    POINTS = GL_POINTS,
    LINE_STRIP = GL_LINE_STRIP
  };

  class Mesh {
    protected:
      GLuint vertexBuffer = -1;
      GLuint indexBuffer = -1;
      GLuint vertexArray = -1;

      /** How many vertices are in the mesh */
      int32_t verticeCount = -1;
      /** How many indices are in the mesh */
      
      int32_t indiceCount = -1;

    public:
      /**
       * Create a new set of buffers for the mesh to use.
       * 
       * @param verticeCount How many Vertices will this buffer support.
       * @param indiceCount How many Indices will this buffer support.
       */
      void createBuffers(
        const int32_t verticeCount,
        const int32_t indiceCount
      );

      /**
       * Cleanup the buffers on a given mesh. This is useful if you intend to
       * expand the count of vertices your mesh supports.
       */
      void disposeBuffers();

      /**
       * Write vertice positions to the mesh.
       * 
       * @param pos Position, within the buffer, to write to.
       * @param positions Array of positions to write.
       * @param len How many positions are in the array.
       */
      void bufferPositions(
        const int32_t pos,
        const glm::vec3 positions[],
        const int32_t len
      );

      /**
       * Write vertice coordinates to the mesh.
       * 
       * @param pos Position, within the buffer, to write to.
       * @param coordinates Array of coordinates to write.
       * @param len How many coordinates are in the array.
       */
      void bufferCoordinates(
        const int32_t pos,
        const glm::vec2 coordinates[],
        const int32_t len
      );

      /**
       * Write indices to the mesh.
       * 
       * @param pos Position, within the buffer, to write to.
       * @param indices Array of indices to write.
       * @param len How many indices are in the array.
       */
      void bufferIndices(
        const int32_t pos,
        const int32_t indices[],
        const int32_t len
      );

      /**
       * Draw a primitive. Primitives are drawn by their indices.
       * 
       * @param drawMode Which drawing mode to use to draw the primitive.
       * @param start Start indice (index) to draw.
       * @param count Count of indices to draw. Use -1 to draw all.
       */
      void draw(
        const enum MeshDrawMode drawMode,
        const int32_t start,
        const int32_t count
      );
      
      /**
       * Cleanup a previously initiated mesh.
       */
      ~Mesh();
  };
}